/*

SampleServer.cpp

This program creates and starts a NatNet server, reads simple marker data from an ascii file,
and streams that data onto the network.  The purpose is to illustrate using the NatNetServer class.

Usage [optional]:

	SampleServer [InputFilename]

	[InputFilename]		Name of points file (.pts) to read from.  Defaults to katie.pts
*/

#include <stdio.h>
#include <tchar.h>
#include <conio.h>
#include <winsock2.h>
#include <stdlib.h>

#include "NatNetTypes.h"
#include "NatNetServer.h"

#pragma warning( disable : 4996 )

void BuildFrame(long FrameNumber, sDataDescriptions* pDataDescriptions, sFrameOfMocapData* pOutFrame);
void SendFrame(sFrameOfMocapData* pFrame);
void FreeFrame(sFrameOfMocapData* pFrame);
void BuildAndSendFrame(long FrameNumber, sDataDescriptions* pDataDescriptions);
int _TrimLine(char *line);
int StartPlayingThread();
int StopPlayingThread();
DWORD WINAPI PlayingThread_Func(void *dummy);
int ReadPTSFile(char* szFile, sDataDescriptions* pDescriptions);
void resetServer();

void __cdecl MessageHandler(int msgType, char* msg);
int __cdecl RequestHandler(sPacket* pPacketIn, sPacket* pPacketOut, void* pUserData);

char szIPAddress[128];
unsigned int Port = 1002;

DWORD PlayingThread_ID = NULL;
HANDLE PlayingThread_Handle = NULL;	
int g_CurrentFrameNumber = -1;		
int nFrames = 0;					// # of frames of data in source file
bool bPlaying = false;				// indicates whether plyaback thread is streaming data

FILE* fp = NULL;					// source file
long lDataStart;					// byte offset in source file where data begins

NatNetServer theServer;				// the NatNet Server 
sDataDescriptions descriptions;

int _tmain(int argc, _TCHAR* argv[])
{
	int retCode;

	// print version info
	unsigned char ver[4];
	theServer.NatNetVersion(ver);
	printf("NatNet Sample Server (NatNet ver. %d.%d.%d.%d)\n", ver[0], ver[1], ver[2], ver[3]);

	// set callbacks
	theServer.SetErrorMessageCallback(MessageHandler);
	theServer.SetVerbosityLevel(Verbosity_Debug);
    theServer.SetMessageResponseCallback( RequestHandler, &theServer );	// this function will receive requests from the Client

	// Initialize server
	unsigned long lAddresses[10];
	int nAddresses;
	nAddresses = theServer.GetLocalIPAddresses(lAddresses, 10);	// get all ip addresses on this machine
	if(nAddresses < 1)
	{
		printf("Unable to detect local network.  Exiting");
		return 1;
	}
	theServer.IPAddress_LongToString(lAddresses[0], szIPAddress);	// use first IP detected
	//strcpy(szIPAddress, "127.0.0.1");
	printf("Starting NatNet server [IP:%s]...\n", szIPAddress);
	//retCode = theServer.Initialize(szIPAddress, Port);
	retCode = theServer.Initialize(szIPAddress);
	if(retCode != 0)
	{
		printf("Error initializing server.  See log for details.  Exiting");
		return 1;
	}

	// Read model definitions (.pts file)
	char szFile[MAX_PATH];
	char szFolder[MAX_PATH];
	GetCurrentDirectory(MAX_PATH, szFolder);
	if(argc > 1)
		sprintf(szFile, "%s\\%s", szFolder, argv[1]);
	else
		sprintf(szFile, "%s\\katie.pts", szFolder);
	printf("Opening data file: %s...\n", szFile);
	nFrames = ReadPTSFile(szFile, &descriptions);
	if(nFrames < 0)
	{
		printf("Error reading file. Exiting");
		return 1;
	}

	// ready to stream data!
	printf("Server started\n\n");
	printf("Commands:\nn\tnext frame\np\tprevious frame\ns\tstream\nr\treset server\nq\tquit\n\n");
	int c;
	bool bExit = false;
	while(c =_getch())
	{
		switch(c)
		{
			case 'n':							// step forward one frame
				g_CurrentFrameNumber++;
				if(g_CurrentFrameNumber >= nFrames)
					g_CurrentFrameNumber = 0;
				BuildAndSendFrame(g_CurrentFrameNumber, &descriptions);
				break;

			case 'p':							// step backward one frame
				g_CurrentFrameNumber--;
				if(g_CurrentFrameNumber < 0)
					g_CurrentFrameNumber = nFrames -1;
				BuildAndSendFrame(g_CurrentFrameNumber, &descriptions);
				break;

			case 's':							// play continuously
				g_CurrentFrameNumber = 0;
				if(bPlaying)
					StopPlayingThread();
				else
					StartPlayingThread();
				break;
			
			case 'q':
				bExit = true;		
				break;
			
			case 'r':	// test : reset server
				resetServer();
				break;
			
			default:
				break;
		}
		if(bExit)
			break;
	}

	fclose(fp);
	return ErrorCode_OK;
}

void resetServer()
{
	int iSuccess = 0;
    
	printf("\nRe-initting Server\n\n");

	iSuccess = theServer.Uninitialize();
	if(iSuccess != 0)
		printf("Error uninitting server.\n");

	iSuccess = theServer.Initialize(szIPAddress, Port);
	if(iSuccess != 0)
		printf("Error re-initting server.\n");

}


// RequestHandler receives requests from the Client.  Server's should
// build and return "request responses" packets in this same thread.
//
//   pPacketIn  - data packet from Client
//   pPacketOut - empty packet, to be filled in and returned to the
//				  Client as the "request response".
//
int __cdecl RequestHandler(sPacket* pPacketIn, sPacket* pPacketOut, void* pUserData)
{
	int iHandled = 1;	// handled

	NatNetServer* pServer = (NatNetServer*) pUserData;	// optional - recover user data set when we setup the callback func


	// Decode the packet into Header and Data
    switch (pPacketIn->iMessage)
    {
    case NAT_PING:
		printf("[SampleServer] received ping from Client.\n");
		printf("[SampleServer] Client App Name : %s\n", pPacketIn->Data.Sender.szName);
		printf("[SampleServer] Client App Version : %d.%d.%d.%d\n", pPacketIn->Data.Sender.Version[0],
				pPacketIn->Data.Sender.Version[1],pPacketIn->Data.Sender.Version[2],pPacketIn->Data.Sender.Version[3]);
		printf("[SampleServer] Client App NatNet Version : %d.%d.%d.%d\n", pPacketIn->Data.Sender.NatNetVersion[0],
				pPacketIn->Data.Sender.NatNetVersion[1],pPacketIn->Data.Sender.NatNetVersion[2],pPacketIn->Data.Sender.NatNetVersion[3]);
		iHandled = 0;
        break;

    case NAT_REQUEST_MODELDEF:
		printf("[SampleServer] Received request for model definition.\n");
		theServer.PacketizeDataDescriptions(&descriptions, pPacketOut);
        break;

	case NAT_REQUEST_FRAMEOFDATA:
		{
		// note: Client does not typically poll for data, but we accomodate it here anyway
		// note: need to return response on same thread as caller, so we cannot call BuildAndSendFrame(...)
		printf("[SampleServer] Received request for frame of data.\n");
		sFrameOfMocapData frame;
		BuildFrame(g_CurrentFrameNumber, &descriptions, &frame);
		theServer.PacketizeFrameOfMocapData(&frame, pPacketOut);
		FreeFrame(&frame);
		}
        break;

	case NAT_REQUEST:

		printf("[SampleServer] Received request from Client: %s\n", pPacketIn->Data.szData);
		pPacketOut->iMessage = NAT_UNRECOGNIZED_REQUEST;

		if (stricmp(pPacketIn->Data.szData, "TestRequest") == 0)
		{
			pPacketOut->iMessage = NAT_RESPONSE;
			strcpy(pPacketOut->Data.szData, "TestResponse");
			pPacketOut->nDataBytes = strlen(pPacketOut->Data.szData) + 1;
		}
		break;

    default:
        pPacketOut->iMessage = NAT_UNRECOGNIZED_REQUEST;
        pPacketOut->nDataBytes = 0;
		iHandled = 0;
        break;
    }

	return iHandled; // 0 = not handled, 1 = handled;
}

// get model description from a simple marker data file (.pts file)
int ReadPTSFile(char* szFile, sDataDescriptions* pDescriptions)
{
	char szMarkerSetName[1024];
	char szMarkerNames[100][1024];		// temp array for holder marker names til we know how many there are
	char szMarkerName[1024];
	char szLine[4096];
	int nMarkerSets = 0;
	int nCharsRead;
	long nFrames = 0;
	int i;

	//todo: modify filereader to read in more than >1 model
	pDescriptions->nDataDescriptions = 1;

	fp = fopen(szFile, "r");
	if(!fp)
		return -1;
	
	sMarkerSetDescription* pMarkerSet = NULL;
	while (!feof(fp))
    {
		fgets(szLine, 1024, fp);
		_TrimLine(szLine);

		if (strcmp(szLine, "<Name>") == 0)									// model's name
        {
			pMarkerSet = new sMarkerSetDescription();
			pDescriptions->arrDataDescriptions[0].type = Descriptor_MarkerSet;
			pDescriptions->arrDataDescriptions[0].Data.MarkerSetDescription = pMarkerSet;
			nCharsRead = fscanf(fp,"%s", szMarkerSetName);
			strcpy(pMarkerSet->szName, szMarkerSetName);
			nMarkerSets++;
        }
        else if (strcmp(szLine, "<Markers>") == 0)							// model's markerset
		{
			nCharsRead = fscanf(fp,"%s", szMarkerName);
			int nMarkers = 0;
			while(strcmp(szMarkerName, "</Markers>")!=0)
			{
				strcpy(szMarkerNames[nMarkers], szMarkerName);
				nMarkers++;
				nCharsRead = fscanf(fp,"%s", szMarkerName);
			}

			pMarkerSet->nMarkers = nMarkers;
			pMarkerSet->szMarkerNames = new char*[nMarkers];
		    for (i = 0; i < nMarkers; ++i)
				pMarkerSet->szMarkerNames[i] = new char[128];
			for(i = 0; i < nMarkers; i++)
			{
				char* szTemp = pMarkerSet->szMarkerNames[i];
                strcpy(szTemp, szMarkerNames[i]);
			}
		}
        else if (strcmp(szLine, "<Data>") == 0)								// data
		{
			// note - we dont read data here, just count the number of frames,
			// *and* preserve the data start line
			// data is actually read during send frame.

			fgets(szLine, 1024, fp);	// skip header line
			lDataStart = ftell(fp);
			fgets(szLine, 4096, fp);
			_TrimLine(szLine);
			while(strcmp(szLine, "</Data>") != 0)
			{
				nFrames++;
				fgets(szLine, 4096, fp);
				_TrimLine(szLine);
			}
		}
	}

	return nFrames;
}

void FreeFrame(sFrameOfMocapData* pFrame)
{
	free(pFrame->MocapData[0].Markers );
	free(pFrame->OtherMarkers);
	for(int i=0; i < pFrame->nRigidBodies; i++)
		free(pFrame->RigidBodies[i].Markers);
}

void BuildFrame(long FrameNumber, sDataDescriptions* pModels, sFrameOfMocapData* pOutFrame)
{

	if(!pModels)
	{
		printf("No models defined - nothing to send.\n");
		return;
	}

	// seek to location in data file
	fseek(fp, lDataStart , SEEK_SET);	// set marker to data start point
	char szLine[4096];
	for(int i=0; i < FrameNumber; i++)
	{
		fgets(szLine, 4096, fp);	// skip frames
	}

	// build frame of data
	sMarkerSetDescription* pMS = pModels->arrDataDescriptions[0].Data.MarkerSetDescription;

	strcpy(pOutFrame->MocapData[0].szName, pMS->szName);
	pOutFrame->iFrame = FrameNumber;
	pOutFrame->nMarkerSets = 1;
	pOutFrame->MocapData[0].nMarkers = pMS->nMarkers;

    // crash fix : must initialize number of unidentified markers:
    pOutFrame->nOtherMarkers =0;

	pOutFrame->fLatency = 66.6f;

	// marker data from file -> frame of data's marker array
	char* ptr;
	ptr = fgets(szLine, 4096, fp);
	long lFrameNumber = atoi(ptr); // should equal FrameNumber
	pOutFrame->MocapData[0].Markers = (MarkerData*)(malloc(pOutFrame->MocapData[0].nMarkers*sizeof(MarkerData)));
	for(int iMarker=0; iMarker < pOutFrame->MocapData[0].nMarkers; iMarker++)
	{	    
		// X value
		ptr++;
		ptr = strchr(ptr, '\t');
		pOutFrame->MocapData[0].Markers[iMarker][0] = (float)atof(ptr);
	
		// Y value
		ptr++;
		ptr = strchr(ptr, '\t');
		pOutFrame->MocapData[0].Markers[iMarker][1] = (float)atof(ptr);

		// Z value
		ptr++;
		ptr = strchr(ptr, '\t');			
		pOutFrame->MocapData[0].Markers[iMarker][2] = (float)atof(ptr);
	}

	// Rigid Bodies
	pOutFrame->nRigidBodies = 10;
	for(int iRigidBody=0; iRigidBody< pOutFrame->nRigidBodies; iRigidBody++)
	{
		pOutFrame->RigidBodies[iRigidBody].ID = iRigidBody;
		pOutFrame->RigidBodies[iRigidBody].x = 3.0f*iRigidBody + 0.0f;
		pOutFrame->RigidBodies[iRigidBody].y = 3.0f*iRigidBody + 0.1f;
		pOutFrame->RigidBodies[iRigidBody].z = 3.0f*iRigidBody + 0.2f;
		pOutFrame->RigidBodies[iRigidBody].qx = 3.0f*iRigidBody + 0.3f;
		pOutFrame->RigidBodies[iRigidBody].qy = 3.0f*iRigidBody + 0.4f;
		pOutFrame->RigidBodies[iRigidBody].qz = 3.0f*iRigidBody + 0.5f;
		pOutFrame->RigidBodies[iRigidBody].qw = 3.0f*iRigidBody + 0.6f;
		pOutFrame->RigidBodies[iRigidBody].nMarkers = 3;

		// markers associated with this rigid body
		pOutFrame->RigidBodies[iRigidBody].Markers = (MarkerData*)(malloc(pOutFrame->RigidBodies[iRigidBody].nMarkers*sizeof(MarkerData)));
		for(int iMarker=0; iMarker < pOutFrame->RigidBodies[iRigidBody].nMarkers; iMarker++)
		{
			pOutFrame->RigidBodies[iRigidBody].Markers[iMarker][0] = iMarker + 0.00f;
			pOutFrame->RigidBodies[iRigidBody].Markers[iMarker][1] = iMarker + 0.01f;
			pOutFrame->RigidBodies[iRigidBody].Markers[iMarker][2] = iMarker + 0.02f;
		}
	}

}

// SendFrame packetizes and sends a single frame of mocap data to the Client
void SendFrame(sFrameOfMocapData* pFrame)
{
	sPacket packet; 
	theServer.PacketizeFrameOfMocapData(pFrame, &packet);
	theServer.SendPacket(&packet);
}

void BuildAndSendFrame(long FrameNumber, sDataDescriptions* pModels)
{
	sFrameOfMocapData frame;
	BuildFrame(FrameNumber, pModels, &frame);
	SendFrame(&frame);
	FreeFrame(&frame);
}

// PlayingThread_Func streams data at ~60hz 
DWORD WINAPI PlayingThread_Func(void * empty)
{
	while (1)
	{
		if(g_CurrentFrameNumber >= nFrames)
			g_CurrentFrameNumber = 0;
		if(g_CurrentFrameNumber < 0)
			g_CurrentFrameNumber = nFrames-1;

		BuildAndSendFrame(g_CurrentFrameNumber, &descriptions);
		printf("Sent Frame %d", g_CurrentFrameNumber);
		Sleep(17);	// ~ 60 fps
		g_CurrentFrameNumber++;
	}

    return ErrorCode_OK;
}

int StartPlayingThread()
{
    SECURITY_ATTRIBUTES security_attribs;
    security_attribs.nLength = sizeof(SECURITY_ATTRIBUTES);
    security_attribs.lpSecurityDescriptor = NULL;
    security_attribs.bInheritHandle = TRUE;

	PlayingThread_Handle = CreateThread(&security_attribs, 0, PlayingThread_Func, NULL, 0, &PlayingThread_ID);
	if(PlayingThread_Handle ==NULL)
	{
		printf("Error creating play thread.");
		return -1;
	}
	else
		bPlaying = true;

	return ErrorCode_OK;
}

int StopPlayingThread()
{
    if (PlayingThread_Handle != NULL)
        TerminateThread(PlayingThread_Handle, 0);
 
    PlayingThread_Handle = NULL;
	bPlaying = false;

	return ErrorCode_OK;
}

// MessageHandler receives NatNet error mesages
void __cdecl MessageHandler(int msgType, char* msg)
{
	printf("\n%s\n", msg);
}

int _TrimLine(char *line)
{
    if (line == NULL)
        return -1;


	// trim ending spaces/characters
    int i = (int) strlen(line);
    while (i > 0)
    {
        if ( (line[i-1] == ' ') || (line[i-1] == '\n') || (line[i-1] == '\r') || (line[i-1] == '\t') )
        {
            line[i-1] = 0;
            i--;
            continue;
        }
        break;
    }

	// trim beginning spaces
    while ( (i > 0)  && ((*line == ' ') || (*line == '\t')) )
    {
        memmove(line, line+1, i);
        i--;
    }

    return i;
}

